import numpy as np
import scipy.stats as ss
import matplotlib.pyplot as plt
import pandas as pd
import stan
import pickle
import seaborn as sb
import scipy.optimize as sciopt
from my_module import *
import arviz as az

   
fin_nm = input('Data file name (*.xlsx) = ')
fout_nm = input('Output file name (*.txt) = ')
fout = open(fout_nm, 'w')
fout.write('Data file...{}\n'.format(fin_nm))
#
#       Read data from Excel file
#
xlsx_fl = pd.ExcelFile(fin_nm)
data_xlsx = pd.read_excel(xlsx_fl,  header = None)
data = data_xlsx.values
var_name = data[0]
values = data[1:]
N = len(values)
ID = []
X = np.empty(N, dtype = float)
Res = np.empty(N, dtype = int)
for t in range(N):
    ID.append(values[t][0])
    X[t] = float(values[t][1])
    Res[t] = int(values[t][2])

print('Number of data = ', N)
fout.write('\nNumber of data = {}\n'.format(N))
for i in range(N):
    print('{0}  {1}  {2}'.format(ID[i], X[i], Res[i]))

for i, s, r in zip(ID, X, Res):
    fout.write('{0:>5}{1:>7.2f}{2:>5d}\n'.format(i, s, r))

#    
#
#       Values for rescaling in MCMC sampling
#
min_sgm = 0.0                       #   lower boundary 
max_sgm = np.amax(X) - np.amin(X)   #   upper boundary 

C1C2 = np.percentile(X, [20, 80])

def f_init():
    """
                Initial values 
    """
    return dict(v_pse = 0.0, v_sgm = 1.0)

stan_data = {'N': N, 'X': X, 'Res': Res, 'min_sgm': min_sgm, 'max_sgm': max_sgm,
             'p75': ss.norm.ppf(0.75, 0, 1) }
#
#       Sampling by Stan
#
with open('PM2CatPS3.stan', 'r') as f_stan:
    sm = stan.build(f_stan.read(), stan_data)
    fit = sm.sample(num_chains = 4, init = [f_init()]*4)

dframe = fit.to_frame()
i_data = az.from_pystan(posterior = fit, posterior_model = sm)

print(az.summary(i_data))
fout.write('\nSummary:\n')
fout.write(f'{az.summary(i_data)}\n')

az.plot_trace(i_data, var_names = ['pse', 'jnd'])
plt.tight_layout()
plt.show()

ck = az.plot_posterior(i_data, var_names = 'pse', point_estimate = 'median')
plt.show()
print(ck)

ck = az.plot_posterior(i_data, var_names = 'jnd', point_estimate = 'median')
plt.show()

#
#       MAP estimates for mu, sigma and JND
#
#
muMAP = CalcMAPEst(dframe['pse'])[0]
sgmMAP = CalcMAPEst(dframe['sgm'])[0]
jndMAP = CalcMAPEst(dframe['jnd'])[0]

SmplSgm = dframe['sgm']   # Samples from the posterior distribution for sigma
SmplPSE = dframe['pse']   # Samples from the posterior distribution for pse(mu)
#
#       Percentile points for posterior distributions
#
SgmL025 = np.percentile(SmplSgm, 2.5)
SgmU975 = np.percentile(SmplSgm, 97.5)
PSEL025 = np.percentile(SmplPSE, 2.5)
PSEU975 = np.percentile(SmplPSE, 97.5)
#
#       Prob(z < s75) = 0.75 for the standard normal distribution
#
s75 = ss.norm.ppf(0.75, 0, 1) 
JND = jndMAP  
PSE = muMAP
fout.write('\nPSE(MAP) = {0:<.3f},    95% CI = [{1:<.2f}, {2:<.2f}]'.format(
            PSE, PSEL025, PSEU975))
fout.write('\nJND(MAP) = {0:<.2f},    95% CI = [{1:<.2f}, {2:<.2f}]'.format(
            JND, s75 * SgmL025, s75 * SgmU975))
#
#       Data and the psychometric function
#
pf = PF(muMAP, sgmMAP)          #   Class PF from my_module.py
minX = X[np.argmin(X)]
maxX = X[np.argmax(X)]
xvalues = []
yvalues = []
for i in range(1001):
    t = minX + i * (maxX - minX) / 1000.0
    xvalues.append(t)
    yvalues.append(pf.v(t))     #   Values of the psychometric function
plt.title('Psychometric Function\n' +
          ('PSE = {0:<.2f},  JND = {1:<.2f}'.format(PSE, JND)),
          fontsize = 16)
plt.plot(xvalues, yvalues, linewidth = 3, label = 'PF')
plt.yticks([0.0, 0.5, 0.75, 1.0])
#
#     Resposes are scaled to values from 0 to 1, and added small fluctuations
#
Res_rn = (Res - 1) + np.random.uniform(-0.05, 0.05, len(Res))   
plt.scatter(X, Res_rn, alpha = 0.5, color = 'g', label = 'Data')
plt.plot([minX, maxX], [0.0, 0.0], color = 'k')
plt.plot([minX, maxX], [1.0, 1.0], color = 'k')
plt.plot([minX, PSE, PSE], [0.5, 0.5, 0.0], color = 'k')
plt.plot([minX, PSE + JND, PSE + JND], [0.75, 0.75, 0.0], color = 'k')
plt.legend(loc = 7)
plt.xlabel('Stimulus value', fontsize = 12)
plt.ylabel('Probability', fontsize = 12)
plt.show()

#
#       Observed responses smoothed by Kernel density estimation 
#
w_1 = np.empty(len(Res))        #   Weights for Response = 2
for t in range(len(Res)):
    w_1[t] = 0 if Res[t] < 2 else 1

X_range = np.arange(minX, maxX, (maxX - minX) / 100)
w_k = JND / 2

while True:
    pm_1 = []           #   KDE values for w_1
    for v in X_range:
        pm_1.append(k_ave(v, X, w_1, w_k))
        
    plt.title('Observed and Estimated PFs\n' +
              ('PSE = {0:<.2f},  JND = {1:<.2f}'.format(PSE, JND)),
              fontsize = 16)
    plt.plot(xvalues, yvalues, color = 'b', linewidth = 3, label = 'PF')
    plt.yticks([0.0, 0.5, 1.0])
    plt.plot(X_range, pm_1, 'g--', linewidth = 2, label = 'Res(>1)')
    plt.xlabel('Stimulus value', fontsize = 12)
    plt.legend(loc = 'upper left')
    plt.show()
    print('\nThe current value of w_k = {}'.format(w_k))
    w_k = float(input('The next value of w_k = '))
    if w_k <= 0.0:
        break

fout.close()

print('\nOutput file {} was saved.'.format(fout_nm))
